from __future__ import annotations

from collections.abc import Generator
from collections.abc import Iterable
from pathlib import Path
from typing import Any
from typing import TypeVar


def get_files_from_dirs(
    dirs: Iterable[Path],
    pattern: str = "*",
) -> Generator[tuple[Path, Path], Any, None]:
    for dir in dirs:
        for path in dir.rglob(pattern):
            if path.is_file():
                yield path, dir


Item = TypeVar("Item")


def unique_ordered(items: Iterable[Item]) -> list[Item]:
    return list(dict.fromkeys(items))



from __future__ import annotations

import re
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any

from django_bird.staticfiles import AssetType


@dataclass
class TestComponent:
    __test__ = False

    name: str
    content: str
    file: Path | None = None
    parent_dir: str = "bird"
    sub_dir: str | None = None

    def create(self, base_dir: Path) -> TestComponent:
        parent = base_dir / self.parent_dir
        parent.mkdir(exist_ok=True)

        if self.sub_dir is not None:
            dir = parent / self.sub_dir
            dir.mkdir(exist_ok=True)
        else:
            dir = parent

        template = dir / f"{self.name}.html"
        template.write_text(self.content)

        self.file = template

        return self


@dataclass
class TestAsset:
    __test__ = False

    component: TestComponent
    content: str
    asset_type: AssetType
    file: Path | None = None

    def create(self) -> TestAsset:
        if self.component.file is None:
            raise ValueError("Component must be created before adding assets")

        component_dir = self.component.file.parent
        component_name = self.component.file.stem

        asset_file = component_dir / f"{component_name}{self.asset_type.ext}"
        asset_file.write_text(self.content)

        self.file = asset_file

        return self


@dataclass
class TestComponentCase:
    __test__ = False

    component: TestComponent
    template_content: str
    expected: str
    description: str = ""
    template_context: dict[str, Any] = field(default_factory=dict)


def print_directory_tree(root_dir: str | Path, prefix: str = ""):
    root_path = Path(root_dir)
    contents = sorted(root_path.iterdir())
    pointers = ["├── "] * (len(contents) - 1) + ["└── "]
    for pointer, path in zip(pointers, contents, strict=False):
        print(prefix + pointer + path.name)
        if path.is_dir():
            extension = "│   " if pointer == "├── " else "    "
            print_directory_tree(path, prefix=prefix + extension)


def normalize_whitespace(text: str) -> str:
    """Normalize whitespace in rendered template output"""
    # multiple whitespace characters
    text = re.sub(r"\s+", " ", text)
    # after opening tag, including when there are attributes
    text = re.sub(r"<(\w+)(\s+[^>]*)?\s*>", r"<\1\2>", text)
    # before closing tag
    text = re.sub(r"\s+>", ">", text)
    # after opening tag and before closing tag
    text = re.sub(r">\s+<", "><", text)
    # immediately after opening tag (including attributes) or before closing tag
    text = re.sub(r"(<\w+(?:\s+[^>]*)?>)\s+|\s+(<\/\w+>)", r"\1\2", text)
    # between tags and text content
    text = re.sub(r">\s+([^<])", r">\1", text)
    text = re.sub(r"([^>])\s+<", r"\1<", text)
    return text.strip()



from __future__ import annotations

import itertools
from collections import defaultdict
from collections.abc import Generator
from collections.abc import Iterable
from dataclasses import dataclass
from dataclasses import field
from hashlib import md5
from pathlib import Path
from threading import Lock
from typing import Any

from cachetools import LRUCache
from django.conf import settings
from django.template.backends.django import Template as DjangoTemplate
from django.template.base import Node
from django.template.base import NodeList
from django.template.base import TextNode
from django.template.context import Context
from django.template.loader import select_template

from .conf import app_settings
from .params import Param
from .params import Params
from .params import Value
from .plugins import pm
from .staticfiles import Asset
from .staticfiles import AssetType
from .templates import gather_bird_tag_template_usage
from .templates import get_component_directories
from .templates import get_template_names
from .templatetags.tags.bird import BirdNode
from .templatetags.tags.slot import DEFAULT_SLOT
from .templatetags.tags.slot import SlotNode
from .utils import get_files_from_dirs


@dataclass(frozen=True, slots=True)
class Component:
    name: str
    template: DjangoTemplate
    assets: frozenset[Asset] = field(default_factory=frozenset)

    def get_asset(self, asset_filename: str) -> Asset | None:
        for asset in self.assets:
            if asset.path.name == asset_filename:
                return asset
        return None

    def get_bound_component(self, node: BirdNode):
        params = Params.from_node(node)
        return BoundComponent(component=self, params=params, nodelist=node.nodelist)

    @property
    def data_attribute_name(self):
        return self.name.replace(".", "-")

    @property
    def id(self):
        normalized_source = "".join(self.source.split())
        hashed = md5(
            f"{self.name}:{self.path}:{normalized_source}".encode()
        ).hexdigest()
        return hashed[:7]

    @property
    def nodelist(self):
        return self.template.template.nodelist

    @property
    def path(self):
        return self.template.template.origin.name

    @property
    def source(self):
        return self.template.template.source

    @classmethod
    def from_abs_path(cls, path: Path, root: Path) -> Component:
        name = str(path.relative_to(root).with_suffix("")).replace("/", ".")
        return cls.from_name(name)

    @classmethod
    def from_name(cls, name: str):
        template_names = get_template_names(name)
        template = select_template(template_names)
        assets: list[Iterable[Asset]] = pm.hook.collect_component_assets(
            template_path=Path(template.template.origin.name)
        )
        return cls(
            name=name, template=template, assets=frozenset(itertools.chain(*assets))
        )


class SequenceGenerator:
    _instance: SequenceGenerator | None = None
    _lock: Lock = Lock()
    _counters: dict[str, int]

    def __init__(self) -> None:
        if not hasattr(self, "_counters"):
            self._counters = {}

    def __new__(cls) -> SequenceGenerator:
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
        return cls._instance

    def next(self, component: Component) -> int:
        with self._lock:
            current = self._counters.get(component.id, 0) + 1
            self._counters[component.id] = current
        return current


@dataclass
class BoundComponent:
    component: Component
    params: Params
    nodelist: NodeList | None
    _sequence: SequenceGenerator = field(default_factory=SequenceGenerator)

    def render(self, context: Context):
        if app_settings.ENABLE_BIRD_ATTRS:
            data_attrs = [
                Param(
                    f"data-bird-{self.component.data_attribute_name}",
                    Value(True),
                ),
                Param("data-bird-id", Value(f'"{self.component.id}-{self.id}"')),
            ]
            self.params.attrs.extend(data_attrs)

        props = self.params.render_props(self.component, context)
        attrs = self.params.render_attrs(context)
        slots = self.fill_slots(context)

        with context.push(
            **{
                "attrs": attrs,
                "props": props,
                "slot": slots.get(DEFAULT_SLOT),
                "slots": slots,
                "vars": {},
            }
        ):
            return self.component.template.template.render(context)

    def fill_slots(self, context: Context):
        if self.nodelist is None:
            return {
                DEFAULT_SLOT: None,
            }

        slot_nodes = {
            node.name: node for node in self.nodelist if isinstance(node, SlotNode)
        }
        default_nodes = NodeList(
            [node for node in self.nodelist if not isinstance(node, SlotNode)]
        )

        slots: dict[str, Node | NodeList] = {
            DEFAULT_SLOT: default_nodes,
            **slot_nodes,
        }

        if not slots[DEFAULT_SLOT] and "slot" in context:
            slots[DEFAULT_SLOT] = TextNode(context["slot"])

        return {name: node.render(context) for name, node in slots.items() if node}

    @property
    def id(self):
        return str(self._sequence.next(self.component))


class ComponentRegistry:
    def __init__(self, maxsize: int = 100):
        self._component_usage: dict[str, set[Path]] = defaultdict(set)
        self._components: LRUCache[str, Component] = LRUCache(maxsize=maxsize)
        self._template_usage: dict[Path, set[str]] = defaultdict(set)

    def discover_components(self) -> None:
        component_dirs = get_component_directories()
        component_paths = get_files_from_dirs(component_dirs)
        for component_abs_path, root_abs_path in component_paths:
            if component_abs_path.suffix != ".html":
                continue

            component = Component.from_abs_path(component_abs_path, root_abs_path)

            if component.name not in self._components:
                self._components[component.name] = component

        templates_using_bird_tag = gather_bird_tag_template_usage()
        for template_abs_path, component_names in templates_using_bird_tag:
            self._template_usage[template_abs_path] = component_names
            for component_name in component_names:
                self._component_usage[component_name].add(template_abs_path)

    def reset(self) -> None:
        """Reset the registry, used for testing."""
        self._component_usage = defaultdict(set)
        self._components.clear()
        self._template_usage = defaultdict(set)

    def get_assets(self, asset_type: AssetType | None = None) -> frozenset[Asset]:
        return frozenset(
            asset
            for component in self._components.values()
            for asset in component.assets
            if asset_type is None or asset.type == asset_type
        )

    def get_component(self, name: str) -> Component:
        if name in self._components and not settings.DEBUG:
            return self._components[name]

        self._components[name] = Component.from_name(name)
        if name not in self._component_usage:
            self._component_usage[name] = set()
        return self._components[name]

    def get_component_usage(
        self, template_path: str | Path
    ) -> Generator[Component, Any, None]:
        path = Path(template_path) if isinstance(template_path, str) else template_path
        for component_name in self._template_usage.get(path, set()):
            yield Component.from_name(component_name)


components = ComponentRegistry()



# pyright: reportAny=false
from __future__ import annotations

from typing import final

from django import template
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context

from django_bird._typing import TagBits
from django_bird._typing import override

TAG = "bird:prop"


def do_prop(_parser: Parser, token: Token) -> PropNode:
    _tag, *bits = token.split_contents()
    if not bits:
        msg = f"{TAG} tag requires at least one argument"
        raise template.TemplateSyntaxError(msg)

    prop = bits.pop(0)

    try:
        name, default = prop.split("=")
    except ValueError:
        name = prop
        default = None

    return PropNode(name, default, bits)


@final
class PropNode(template.Node):
    def __init__(self, name: str, default: str | None, attrs: TagBits):
        self.name = name
        self.default = default
        self.attrs = attrs

    @override
    def render(self, context: Context) -> str:
        return ""



from __future__ import annotations

import contextlib
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from django.conf import settings
from django.template.backends.django import DjangoTemplates
from django.template.backends.django import Template as DjangoTemplate
from django.template.engine import Engine
from django.test import override_settings
from django.urls import clear_url_caches
from django.urls import include
from django.urls import path

from .settings import DEFAULT_SETTINGS

if TYPE_CHECKING:
    from django_bird.components import Component

SLOW_MARK = "slow"
SLOW_CLI_ARG = "--slow"

pytest_plugins = []


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(SLOW_CLI_ARG, action="store_true", help="run tests marked as slow")


def pytest_configure(config: pytest.Config) -> None:  # pyright: ignore [reportUnusedParameter]
    logging.disable(logging.CRITICAL)

    settings.configure(**DEFAULT_SETTINGS, **TEST_SETTINGS)


def pytest_runtest_setup(item: pytest.Item) -> None:
    if SLOW_MARK in item.keywords and not item.config.getoption(SLOW_CLI_ARG):
        pytest.skip(f"pass {SLOW_CLI_ARG} to run slow tests")


TEST_SETTINGS = {
    "INSTALLED_APPS": [
        "django_bird",
        "django.contrib.staticfiles",
    ],
    "STATIC_URL": "/static/",
    "STATICFILES_FINDERS": [
        "django.contrib.staticfiles.finders.FileSystemFinder",
        "django.contrib.staticfiles.finders.AppDirectoriesFinder",
        "django_bird.staticfiles.BirdAssetFinder",
    ],
    "TEMPLATES": [
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "DIRS": [
                Path(__file__).parent / "templates",
            ],
            "OPTIONS": {
                "builtins": [
                    "django.template.defaulttags",
                    "django_bird.templatetags.django_bird",
                ],
                "loaders": [
                    "django.template.loaders.filesystem.Loader",
                    "django.template.loaders.app_directories.Loader",
                ],
            },
        }
    ],
}


@pytest.fixture(autouse=True)
def setup_urls():
    urlpatterns = [
        path("__bird__/", include("django_bird.urls")),
    ]

    clear_url_caches()

    with override_settings(
        ROOT_URLCONF=type(
            "urls",
            (),
            {"urlpatterns": urlpatterns},
        ),
    ):
        yield

    clear_url_caches()


@pytest.fixture
def templates_dir(tmp_path):
    templates_dir = tmp_path / "templates"
    templates_dir.mkdir()
    return templates_dir


@pytest.fixture(autouse=True)
def override_templates_settings(templates_dir):
    with override_settings(
        TEMPLATES=[
            settings.TEMPLATES[0]
            | {
                "DIRS": [
                    *settings.TEMPLATES[0]["DIRS"],
                    templates_dir,
                ]
            }
        ]
    ):
        yield


@pytest.fixture
def override_app_settings():
    from django_bird.conf import DJANGO_BIRD_SETTINGS_NAME

    @contextlib.contextmanager
    def _override_app_settings(**kwargs):
        with override_settings(**{DJANGO_BIRD_SETTINGS_NAME: {**kwargs}}):
            yield

    return _override_app_settings


@pytest.fixture(autouse=True)
def data_bird_attr_app_setting(override_app_settings, request):
    enable = "default_app_settings" in request.keywords

    with override_app_settings(ENABLE_BIRD_ATTRS=enable):
        yield


@pytest.fixture
def create_template():
    def _create_template(template_file: Path) -> DjangoTemplate:
        engine = Engine(
            builtins=["django_bird.templatetags.django_bird"],
            dirs=[str(template_file.parent)],
        )
        template = engine.get_template(template_file.name)
        backend = DjangoTemplates(
            {
                "NAME": "django",
                "DIRS": [],
                "APP_DIRS": True,
                "OPTIONS": {
                    "autoescape": True,
                    "debug": False,
                    "context_processors": [],
                },
            }
        )
        return DjangoTemplate(template, backend)

    return _create_template


@dataclass
class ExampleTemplate:
    base: Path
    include: Path
    template: Path
    used_components: list[Component]
    unused_components: list[Component]

    @property
    def content(self):
        return self.template.read_text()


@pytest.fixture
def example_template(templates_dir):
    from django_bird.components import Component
    from django_bird.staticfiles import AssetType

    from .utils import TestAsset
    from .utils import TestComponent

    button = TestComponent(name="button", content="<button>{{ slot }}</button>").create(
        templates_dir
    )
    TestAsset(
        component=button,
        content=".button { color: blue; }",
        asset_type=AssetType.CSS,
    ).create()
    TestAsset(
        component=button, content="console.log('button');", asset_type=AssetType.JS
    ).create()

    alert = TestComponent(
        name="alert", content='<div class="alert">{{ slot }}</div>'
    ).create(templates_dir)
    TestAsset(
        component=alert, content=".alert { color: red; }", asset_type=AssetType.CSS
    ).create()

    banner = TestComponent(
        name="banner", content="<div {{ attrs }}>{{ slot }}</div>"
    ).create(templates_dir)

    toast = TestComponent(name="toast", content="<div>{{ slot }}</div>").create(
        templates_dir
    )
    TestAsset(
        component=toast,
        content=".toast { color: pink; }",
        asset_type=AssetType.CSS,
    ).create()
    TestAsset(
        component=toast, content="console.log('toast');", asset_type=AssetType.JS
    ).create()

    base_template = templates_dir / "base.html"
    base_template.write_text("""
        <html>
        <head>
            <title>Test</title>
            {% bird:css %}
        </head>
        <body>
            {% bird alert %}Warning{% endbird %}
            {% block content %}{% endblock %}
            {% bird:js %}
        </body>
        </html>
    """)

    include_template = templates_dir / "include.html"
    include_template.write_text("""
        {% bird banner %}Include me{% endbird %}
    """)

    template = templates_dir / "template.html"
    template.write_text("""
        {% extends "base.html" %}
        {% block content %}
            {% include "include.html" %}
            {% bird button %}Click me{% endbird %}
        {% endblock %}
    """)

    return ExampleTemplate(
        base=base_template,
        include=include_template,
        template=template,
        used_components=[
            Component.from_name(alert.name),
            Component.from_name(banner.name),
            Component.from_name(button.name),
        ],
        unused_components=[Component.from_name(toast.name)],
    )


@pytest.fixture(autouse=True)
def registry():
    from django_bird.components import components

    components.reset()
    yield components
    components.reset()



# pyright: reportAny=false
from __future__ import annotations

from typing import cast
from typing import final

from django import template
from django.template.base import NodeList
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context
from django.utils.safestring import SafeString

from django_bird._typing import override

TAG = "bird:slot"
END_TAG = "endbird:slot"

DEFAULT_SLOT = "default"


def do_slot(parser: Parser, token: Token) -> SlotNode:
    _tag, *bits = token.split_contents()
    if len(bits) > 1:
        msg = f"{TAG} tag requires either one or no arguments"
        raise template.TemplateSyntaxError(msg)

    if len(bits) == 0:
        name = DEFAULT_SLOT
    else:
        name = bits[0]
        if name.startswith("name="):
            _, name = name.split("=")
        name = name.strip("'\"")

    nodelist = parser.parse((END_TAG,))
    parser.delete_first_token()

    return SlotNode(name, nodelist)


@final
class SlotNode(template.Node):
    def __init__(self, name: str, nodelist: NodeList):
        self.name = name
        self.nodelist = nodelist

    @override
    def render(self, context: Context) -> SafeString:
        slots = context.get("slots")

        if not slots or not isinstance(slots, dict):
            return self.nodelist.render(context)

        slots_dict = cast(dict[str, str], slots)
        slot_content = slots_dict.get(self.name)

        if slot_content is None or slot_content == "":
            return self.nodelist.render(context)

        return template.Template(slot_content).render(context)



from __future__ import annotations

import pytest

from django_bird.utils import get_files_from_dirs
from django_bird.utils import unique_ordered

from .utils import normalize_whitespace


class TestBirdUtils:
    def test_get_files_from_dirs(self, tmp_path):
        first = tmp_path / "first"
        first.mkdir()

        for i in range(10):
            file = first / f"to_find{i}.txt"
            file.write_text("file should be found")

        for i in range(10):
            file = first / f"do_not_find{i}.txt"
            file.write_text("file should not be found")

        second = tmp_path / "second"
        second.mkdir()

        for i in range(10):
            file = second / f"to_find{i}.txt"
            file.write_text("file should be found")

        for i in range(10):
            file = second / f"do_not_find{i}.txt"
            file.write_text("file should not be found")

        dirs = [first, second]

        found_paths = list(get_files_from_dirs(dirs, "to_find*.txt"))

        assert len(found_paths) == 20
        assert all("to_find" in path.name for path, _ in found_paths)
        assert not any("do_not_find" in path.name for path, _ in found_paths)

    @pytest.mark.parametrize(
        "items,expected",
        [
            (["a", "b", "c", "a"], ["a", "b", "c"]),
            (["first", "second", "first", "third"], ["first", "second", "third"]),
            ([1, 2, 1, 3], [1, 2, 3]),
            ([1, "b", 1, "a"], [1, "b", "a"]),
        ],
    )
    def test_unique_ordered(self, items, expected):
        assert unique_ordered(items) == expected


class TestTestsUtils:
    @pytest.mark.parametrize(
        "contents,expected",
        [
            ("<button> Click me </button>", "<button>Click me</button>"),
            ("<button> Click me</button>", "<button>Click me</button>"),
            ("<button>Click me </button>", "<button>Click me</button>"),
            ("<button >Click me</button>", "<button>Click me</button>"),
            (
                "<button class='btn'> Click me </button>",
                "<button class='btn'>Click me</button>",
            ),
            ("\n<button>\n  Click me\n</button>\n", "<button>Click me</button>"),
        ],
    )
    def test_normalize_whitespace(self, contents, expected):
        assert normalize_whitespace(contents) == expected



from __future__ import annotations

import pytest
from django import template
from django.template import Context
from django.template import Template
from django.template.base import Parser
from django.template.base import Token
from django.template.base import TokenType
from django.template.exceptions import TemplateSyntaxError

from django_bird.templatetags.tags.var import END_TAG
from django_bird.templatetags.tags.var import TAG
from django_bird.templatetags.tags.var import do_end_var
from django_bird.templatetags.tags.var import do_var
from tests.utils import TestComponent


def test_do_var_no_args():
    start_token = Token(TokenType.BLOCK, TAG)

    with pytest.raises(template.TemplateSyntaxError):
        do_var(Parser([]), start_token)


def test_do_end_var_no_args():
    start_token = Token(TokenType.BLOCK, END_TAG)

    with pytest.raises(template.TemplateSyntaxError):
        do_end_var(Parser([]), start_token)


def test_basic_assignment():
    template = Template("""
        {% load django_bird %}
        {% bird:var x='hello' %}
        {{ vars.x }}
    """)

    rendered = template.render(Context({}))

    assert rendered.strip() == "hello"


def test_append_to_variable():
    template = Template("""
        {% load django_bird %}
        {% bird:var x='hello' %}
        {% bird:var x+=' world' %}
        {{ vars.x }}
    """)

    rendered = template.render(Context({}))

    assert rendered.strip() == "hello world"


def test_multiple_variables():
    template = Template("""
        {% load django_bird %}
        {% bird:var x='hello' %}
        {% bird:var y=' world' %}
        {{ vars.x }}{{ vars.y }}
    """)

    rendered = template.render(Context({}))

    assert rendered.strip() == "hello world"


def test_append_to_nonexistent_variable():
    template = Template("""
        {% load django_bird %}
        {% bird:var x+='world' %}
        {{ vars.x }}
    """)

    rendered = template.render(Context({}))

    assert rendered.strip() == "world"


def test_variable_with_template_variable():
    template = Template("""
        {% load django_bird %}
        {% bird:var greeting='Hello ' %}
        {% bird:var greeting+=name %}
        {{ vars.greeting }}
    """)

    rendered = template.render(Context({"name": "Django"}))

    assert rendered.strip() == "Hello Django"


def test_explicit_var_cleanup():
    template = Template("""
        {% load django_bird %}
        {% bird:var x='hello' %}
        Before: {{ vars.x }}
        {% endbird:var x %}
        After: {{ vars.x|default:'cleaned' }}
    """)

    rendered = template.render(Context({}))

    assert "Before: hello" in rendered
    assert "After: cleaned" in rendered


def test_reseting_var():
    template = Template("""
        {% load django_bird %}
        {% bird:var x='hello' %}
        Before: {{ vars.x }}
        {% bird:var x=None %}
        After: {{ vars.x|default:'cleaned' }}
    """)

    rendered = template.render(Context({}))

    assert "Before: hello" in rendered
    assert "After: cleaned" in rendered


def test_explicit_var_cleanup_with_multiple_vars():
    template = Template("""
        {% load django_bird %}
        {% bird:var x='hello' %}
        {% bird:var y='world' %}
        Before: {{ vars.x }} {{ vars.y }}
        {% endbird:var x %}
        Middle: {{ vars.x|default:'cleaned' }} {{ vars.y }}
        {% endbird:var y %}
        After: {{ vars.x|default:'cleaned' }} {{ vars.y|default:'cleaned' }}
    """)

    rendered = template.render(Context({}))

    assert "Before: hello world" in rendered
    assert "Middle: cleaned world" in rendered
    assert "After: cleaned cleaned" in rendered


@pytest.mark.parametrize(
    ("template_str", "expected_error"),
    [
        (
            "{% load django_bird %}{% bird:var %}",
            r"\'bird:var\' tag requires an assignment",
        ),
        (
            "{% load django_bird %}{% bird:var invalid_syntax %}",
            r"Invalid assignment in \'bird:var\' tag: invalid_syntax\. Expected format: bird:var variable=\'value\' or bird:var variable\+=\'value\'\.",
        ),
    ],
)
def test_syntax_errors(template_str: str, expected_error: str):
    with pytest.raises(TemplateSyntaxError, match=expected_error):
        Template(template_str)


def test_var_context_isolation_between_components(create_template, templates_dir):
    TestComponent(
        name="button",
        content="""
            {% bird:var x='button1' %}
            {{ vars.x }}
        """,
    ).create(templates_dir)

    template_path = templates_dir / "test.html"
    template_path.write_text("""
        {% bird button %}{% endbird %}
        {% bird button %}{% endbird %}
    """)
    template = create_template(template_path)

    rendered = template.render({})

    assert "button1" in rendered
    assert rendered.count("button1") == 2


def test_var_context_does_not_leak_outside_component(create_template, templates_dir):
    TestComponent(
        name="button",
        content="""
            {% bird:var x='button1' %}
            Inner: {{ vars.x }}
        """,
    ).create(templates_dir)

    template_path = templates_dir / "test.html"
    template_path.write_text("""
        {% bird button %}{% endbird %}
        Outer: {{ vars.x|default:'not accessible' }}
    """)
    template = create_template(template_path)

    rendered = template.render({})

    assert "Inner: button1" in rendered
    assert "Outer: not accessible" in rendered


def test_var_context_isolation_nested_components(create_template, templates_dir):
    TestComponent(
        name="outer",
        content="""
            {% bird:var x='outer' %}
            Outer: {{ vars.x }}
            {% bird inner %}{% endbird %}
            After Inner: {{ vars.x }}
        """,
    ).create(templates_dir)
    TestComponent(
        name="inner",
        content="""
            {% bird:var x='inner' %}
            Inner: {{ vars.x }}
        """,
    ).create(templates_dir)

    template_path = templates_dir / "test.html"
    template_path.write_text("""
        {% bird outer %}{% endbird %}
    """)
    template = create_template(template_path)

    rendered = template.render({})

    assert "Outer: outer" in rendered
    assert "Inner: inner" in rendered
    assert "After Inner: outer" in rendered


def test_var_context_clean_between_renders(create_template, templates_dir):
    TestComponent(
        name="counter",
        content="""
            {% bird:var count='1' %}
            Count: {{ vars.count }}
        """,
    ).create(templates_dir)

    template_path = templates_dir / "test.html"
    template_path.write_text("{% bird counter %}{% endbird %}")
    template = create_template(template_path)

    first_render = template.render({})
    second_render = template.render({})

    assert "Count: 1" in first_render
    assert "Count: 1" in second_render


def test_var_append_with_nested_components(create_template, templates_dir):
    TestComponent(
        name="outer",
        content="""
            {% bird:var message='Hello' %}
            {% bird:var message+=' outer' %}
            Outer: {{ vars.message }}
            {% bird inner %}{% endbird %}
            After: {{ vars.message }}
        """,
    ).create(templates_dir)
    TestComponent(
        name="inner",
        content="""
            {% bird:var message='Hello' %}
            {% bird:var message+=' inner' %}
            Inner: {{ vars.message }}
        """,
    ).create(templates_dir)

    template_path = templates_dir / "test.html"
    template_path.write_text("{% bird outer %}{% endbird %}")
    template = create_template(template_path)

    rendered = template.render({})

    assert "Outer: Hello outer" in rendered
    assert "Inner: Hello inner" in rendered
    assert "After: Hello outer" in rendered



from __future__ import annotations

import shutil
from pathlib import Path

import pytest
from django.contrib.staticfiles import finders
from django.core.management import call_command
from django.template.base import Template
from django.template.context import Context
from django.test import override_settings

from django_bird.components import Component
from django_bird.staticfiles import Asset
from django_bird.staticfiles import AssetType
from django_bird.staticfiles import BirdAssetFinder
from django_bird.staticfiles import collect_component_assets

from .utils import TestAsset
from .utils import TestComponent


class TestAssetClass:
    def test_hash(self):
        asset1 = Asset(Path("static.css"), AssetType.CSS)
        asset2 = Asset(Path("static.css"), AssetType.CSS)

        assert asset1 == asset2
        assert hash(asset1) == hash(asset2)

        assets = {asset1, asset2, Asset(Path("other.css"), AssetType.CSS)}

        assert len(assets) == 2

    def test_exists(self, tmp_path: Path):
        css_file = tmp_path / "test.css"
        css_file.touch()

        asset = Asset(css_file, AssetType.CSS)

        assert asset.exists() is True

    def test_exists_nonexistent(self):
        missing_asset = Asset(Path("missing.css"), AssetType.CSS)

        assert missing_asset.exists() is False

    @pytest.mark.parametrize(
        "asset,expected_html_tag_bits",
        [
            (
                TestAsset(
                    component=None,
                    content=".button { color: blue; }",
                    asset_type=AssetType.CSS,
                ),
                'link rel="stylesheet" href=',
            ),
            (
                TestAsset(
                    component=None,
                    content="console.log('button');",
                    asset_type=AssetType.JS,
                ),
                "script src=",
            ),
        ],
    )
    def test_render(self, asset, expected_html_tag_bits, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        asset.component = button
        asset.create()

        component = Component.from_name(button.name)
        component_asset = component.get_asset(asset.file.name)

        rendered = component_asset.render()

        assert expected_html_tag_bits in rendered
        assert component_asset.url in rendered

    def test_storage(self, templates_dir, settings):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        component = Component.from_name(button.name)
        asset = component.get_asset(button_css.file.name)

        with override_settings(
            STATIC_URL="/static/",
        ):
            assert str(asset.storage.location) in str(button_css.file)

    def test_template_dir(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        component = Component.from_name(button.name)
        asset = component.get_asset(button_css.file.name)

        assert asset.template_dir == templates_dir

    def test_template_dir_nested(self, templates_dir):
        button = TestComponent(
            name="button",
            content="<button>Click me</button>",
            sub_dir="nested",
        ).create(templates_dir)
        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        component = Component.from_name(f"{button.sub_dir}.{button.name}")
        asset = component.get_asset(button_css.file.name)

        assert asset.template_dir == templates_dir

    def test_relative_path(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        component = Component.from_name(button.name)
        asset = component.get_asset(button_css.file.name)

        assert asset.relative_path == Path("bird/button.css")

    def test_relative_path_nested(self, templates_dir):
        button = TestComponent(
            name="button",
            content="<button>Click me</button>",
            sub_dir="nested",
        ).create(templates_dir)
        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        component = Component.from_name(f"{button.sub_dir}.{button.name}")
        asset = component.get_asset(button_css.file.name)

        assert asset.relative_path == Path("bird/nested/button.css")

    def test_url(self, templates_dir):
        button = TestComponent(
            name="button",
            content="<button>Click me</button>",
        ).create(templates_dir)
        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        component = Component.from_name(button.name)
        asset = component.get_asset(button_css.file.name)

        assert (
            asset.url == f"/static/{button_css.file.parent.name}/{button_css.file.name}"
        )

    def test_url_nonexistent(self, templates_dir):
        button = TestComponent(
            name="button",
            content="<button>Click me</button>",
        ).create(templates_dir)
        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        component = Component.from_name(button.name)
        asset = component.get_asset(button_css.file.name)

        button_css.file.unlink()

        assert asset.url is None


def test_asset_collection(templates_dir):
    button = TestComponent(name="button", content="<button>Click me</button>").create(
        templates_dir
    )
    button_css = TestAsset(
        component=button,
        content=".button { color: blue; }",
        asset_type=AssetType.CSS,
    ).create()
    button_js = TestAsset(
        component=button, content="console.log('button');", asset_type=AssetType.JS
    ).create()

    assets = collect_component_assets(button.file)

    assert Asset(button_css.file, button_css.asset_type) in assets
    assert Asset(button_js.file, button_js.asset_type) in assets


def test_asset_collection_nested(templates_dir):
    button = TestComponent(
        name="button", content="<button>Click me</button>", sub_dir="nested"
    ).create(templates_dir)
    button_css = TestAsset(
        component=button,
        content=".button { color: blue; }",
        asset_type=AssetType.CSS,
    ).create()

    assets = collect_component_assets(button.file)

    assert Asset(button_css.file, button_css.asset_type) in assets


class TestBirdAssetFinder:
    def test_check(self):
        finder = BirdAssetFinder()

        assert finder.check() == []

    def test_list_all_assets(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        input = TestComponent(name="input", content="<input>").create(templates_dir)

        assets = [
            TestAsset(
                component=button,
                content=".button { color: blue; }",
                asset_type=AssetType.CSS,
            ).create(),
            TestAsset(
                component=button,
                content="console.log('button');",
                asset_type=AssetType.JS,
            ).create(),
            TestAsset(
                component=input,
                content="input { border: 1px solid black; }",
                asset_type=AssetType.CSS,
            ).create(),
            TestAsset(
                component=input,
                content="console.log('input');",
                asset_type=AssetType.JS,
            ).create(),
        ]

        finder = BirdAssetFinder()
        listed_assets = list(finder.list(None))

        assert len(listed_assets) == len(assets)

        for asset in assets:
            assert any(
                asset.file.name in relative_path for relative_path, _ in listed_assets
            )

    def test_list_ignore(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        finder = BirdAssetFinder()

        assert len(list(finder.list(["bird/*"]))) == 0
        assert len(list(finder.list(["*/button*"]))) == 0
        assert len(list(finder.list(["*.css"]))) == 0

    def test_list_custom_dir(self, templates_dir, override_app_settings):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        custom_button = TestComponent(
            name="button", content="<button>Click me</button>", parent_dir="components"
        ).create(templates_dir)

        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()
        custom_button_css = TestAsset(
            component=custom_button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        finder = BirdAssetFinder()

        with override_app_settings(COMPONENT_DIRS=["components"]):
            listed_assets = list(finder.list(None))

        assert len(listed_assets) == 1
        # components are matched like templates, dirs in
        # `settings.DJANGO_BIRD["COMPONENT_DIRS"]` take precedence
        assert listed_assets[0][0] in str(custom_button_css.file)
        assert listed_assets[0][0] not in str(button_css.file)


class TestFindersFind:
    def test_find_asset(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()
        button_js = TestAsset(
            component=button, content="console.log('button');", asset_type=AssetType.JS
        ).create()

        assert finders.find("bird/button.css") == str(button_css.file)
        assert finders.find("bird/button.js") == str(button_js.file)

    def test_find_asset_custom_dir(self, templates_dir, override_app_settings):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        custom_button = TestComponent(
            name="button", content="<button>Click me</button>", parent_dir="components"
        ).create(templates_dir)

        TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()
        custom_button_css = TestAsset(
            component=custom_button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        with override_app_settings(COMPONENT_DIRS=["components"]):
            # components are matched like templates, dirs in
            # `settings.DJANGO_BIRD["COMPONENT_DIRS"]` take precedence
            assert finders.find("components/button.css") == str(custom_button_css.file)
            assert finders.find("bird/button.css") is None

    def test_find_nested_asset(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>", sub_dir="nested"
        ).create(templates_dir)

        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        assert finders.find("bird/nested/button.css") == str(button_css.file)

    def test_case_sensitive_component_names(self, templates_dir):
        component = TestComponent(name="Button", content="<button>").create(
            templates_dir
        )
        TestAsset(component=component, content="", asset_type=AssetType.CSS).create()

        assert finders.find("bird/Button.css") is not None
        assert finders.find("bird/button.css") is None

    def test_find_intermediate_paths(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>", sub_dir="nested"
        ).create(templates_dir)

        TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        assert finders.find("bird/") is not None
        assert finders.find("bird/nested/") is not None
        assert finders.find("bird/nested/button.css") is not None

    def test_find_component_template(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        assert finders.find("bird/button.html") is None

    def test_find_nonexistent(self):
        assert finders.find("bird/nonexistent.js") is None

    def test_find_all(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        assert finders.find("bird/button.css", all=True) == [str(button_css.file)]


class TestStaticCollection:
    @pytest.fixture
    def static_root(self, tmp_path):
        static_dir = tmp_path / "static"
        static_dir.mkdir()

        with override_settings(STATIC_ROOT=str(static_dir)):
            yield static_dir

        shutil.rmtree(static_dir)

    def test_basic(self, templates_dir, static_root):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()
        TestAsset(
            component=button,
            content="console.log('button');",
            asset_type=AssetType.JS,
        ).create()

        call_command("collectstatic", interactive=False, verbosity=0)

        assert (static_root / "django_bird/bird/button.css").exists()
        assert (static_root / "django_bird/bird/button.js").exists()

    def test_custom_dir(self, templates_dir, static_root, override_app_settings):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        custom_button = TestComponent(
            name="button", content="<button>Click me</button>", parent_dir="components"
        ).create(templates_dir)

        TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()
        TestAsset(
            component=custom_button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        with override_app_settings(COMPONENT_DIRS=["components"]):
            call_command("collectstatic", interactive=False, verbosity=0)

            # components are matched like templates, dirs in
            # `settings.DJANGO_BIRD["COMPONENT_DIRS"]` take precedence
            assert (static_root / "django_bird/components/button.css").exists()
            assert not (static_root / "django_bird/bird/button.css").exists()

    def test_same_named_assets(self, templates_dir, static_root):
        button1 = TestComponent(
            name="button", content="<button>Form Button</button>", sub_dir="forms"
        ).create(templates_dir)
        button2 = TestComponent(
            name="button", content="<button>Nav Button</button>", sub_dir="nav"
        ).create(templates_dir)

        button1_css = TestAsset(
            component=button1,
            content=".form-button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()
        button2_css = TestAsset(
            component=button2,
            content=".nav-button { color: red; }",
            asset_type=AssetType.CSS,
        ).create()

        call_command("collectstatic", interactive=False, verbosity=0)

        form_css = static_root / "django_bird/bird/forms/button.css"
        nav_css = static_root / "django_bird/bird/nav/button.css"

        assert form_css.exists()
        assert nav_css.exists()
        assert form_css.read_text() == button1_css.content
        assert nav_css.read_text() == button2_css.content
        assert form_css.read_text() != button2_css.content
        assert nav_css.read_text() != button1_css.content

    def test_dry_run_collection(self, templates_dir, static_root):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        call_command("collectstatic", interactive=False, dry_run=True, verbosity=0)

        assert not (static_root / "bird").exists()

    def test_clear_collection(self, templates_dir, static_root):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        call_command("collectstatic", interactive=False, verbosity=0)

        assert (static_root / "django_bird/bird/button.css").exists()
        assert not (static_root / "django_bird/bird/input.css").exists()

        input = TestComponent(name="input", content="<input>").create(templates_dir)
        TestAsset(
            component=input,
            content="input { border: 1px solid black; }",
            asset_type=AssetType.CSS,
        ).create()

        call_command("collectstatic", interactive=False, clear=True, verbosity=0)

        assert (static_root / "django_bird/bird/button.css").exists()
        assert (static_root / "django_bird/bird/input.css").exists()


class TestStaticTemplateTag:
    def test_static_tag(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)
        TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        template = Template(
            '{% load static %}{% static "django_bird/bird/button.css" %}'
        )
        rendered = template.render(Context({}))

        assert rendered == "/static/django_bird/bird/button.css"

    def test_static_tag_nonexistent_asset(self):
        template = Template(
            '{% load static %}{% static "django_bird/bird/nonexistent.css" %}'
        )
        rendered = template.render(Context({}))

        assert rendered == "/static/django_bird/bird/nonexistent.css"



from __future__ import annotations

import pytest
from django import template
from django.template.base import Parser
from django.template.base import Token
from django.template.base import TokenType

from django_bird.templatetags.tags.prop import TAG
from django_bird.templatetags.tags.prop import PropNode
from django_bird.templatetags.tags.prop import do_prop


@pytest.mark.parametrize(
    "contents,expected",
    [
        ("id", PropNode(name="id", default=None, attrs=[])),
        ("class='btn'", PropNode(name="class", default="'btn'", attrs=[])),
    ],
)
def test_do_prop(contents, expected):
    start_token = Token(TokenType.BLOCK, f"{TAG} {contents}")

    node = do_prop(Parser([]), start_token)

    assert node.name == expected.name
    assert node.default == expected.default
    assert node.attrs == expected.attrs


def test_do_prop_no_args():
    start_token = Token(TokenType.BLOCK, TAG)

    with pytest.raises(template.TemplateSyntaxError):
        do_prop(Parser([]), start_token)



from __future__ import annotations

import pytest
from django.template import Context
from django.template import Template
from django.template.base import Parser
from django.template.base import Token
from django.template.base import TokenType
from django.template.exceptions import TemplateSyntaxError

from django_bird.templatetags.tags.slot import DEFAULT_SLOT
from django_bird.templatetags.tags.slot import END_TAG
from django_bird.templatetags.tags.slot import TAG
from django_bird.templatetags.tags.slot import SlotNode
from django_bird.templatetags.tags.slot import do_slot
from tests.utils import TestComponent
from tests.utils import TestComponentCase
from tests.utils import normalize_whitespace


@pytest.mark.parametrize(
    "contents,expected",
    [
        ("", SlotNode(name=DEFAULT_SLOT, nodelist=None)),
        ("foo", SlotNode(name="foo", nodelist=None)),
        ("'foo'", SlotNode(name="foo", nodelist=None)),
        ('"foo"', SlotNode(name="foo", nodelist=None)),
        ('name="foo"', SlotNode(name="foo", nodelist=None)),
        ("name='foo'", SlotNode(name="foo", nodelist=None)),
    ],
)
def test_parse_slot_name(contents, expected):
    start_token = Token(TokenType.BLOCK, f"{TAG} {contents}")
    end_token = Token(TokenType.BLOCK, END_TAG)

    node = do_slot(Parser([end_token]), start_token)

    assert node.name == expected.name


class TestTemplateTag:
    @pytest.mark.parametrize(
        "component_content",
        [
            "{{ slot }}",
            "{% bird:slot %}{% endbird:slot %}",
            "{% bird:slot default %}{% endbird:slot %}",
            "{% bird:slot 'default' %}{% endbird:slot %}",
            '{% bird:slot "default" %}{% endbird:slot %}',
            "{% bird:slot name=default %}{% endbird:slot %}",
            "{% bird:slot name='default' %}{% endbird:slot %}",
            '{% bird:slot name="default" %}{% endbird:slot %}',
        ],
    )
    def test_default_slot(self, component_content, templates_dir):
        test_case = TestComponentCase(
            component=TestComponent(
                name="test",
                content=component_content,
            ),
            template_content="{% bird test %}Content{% endbird %}",
            template_context={"slot": "Content"},
            expected="Content",
        )
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    def test_default_content(self, templates_dir):
        test_case = TestComponentCase(
            component=TestComponent(
                name="test",
                content="""
                    <button>
                        {% bird:slot leading-icon %}
                            <span>Default icon</span>
                        {% endbird:slot %}

                        {% bird:slot %}
                            Click me
                        {% endbird:slot %}
                    </button>
                """,
            ),
            template_content="{% bird test %}{% endbird %}",
            expected="<button><span>Default icon</span>Click me</button>",
        )
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context({}))

        assert normalize_whitespace(rendered) == test_case.expected

    def test_default_content_override(self, templates_dir):
        test_case = TestComponentCase(
            component=TestComponent(
                name="test",
                content="""
                    <button>
                        {% bird:slot leading-icon %}
                            <span>Default icon</span>
                        {% endbird:slot %}

                        {% bird:slot %}
                            Click me
                        {% endbird:slot %}
                    </button>
                """,
            ),
            template_content="""
                {% bird test %}
                    {% bird:slot leading-icon %}→{% endbird:slot %}
                    Submit
                {% endbird %}
            """,
            expected="<button>→ Submit</button>",
        )
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context({}))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "component_content",
        [
            "{{ slots.named_slot }}",
            "{% bird:slot named_slot %}{% endbird:slot %}",
            "{% bird:slot name=named_slot %}{% endbird:slot %}",
        ],
    )
    def test_named_slot(self, component_content, templates_dir):
        test_case = TestComponentCase(
            component=TestComponent(
                name="test",
                content=component_content,
            ),
            template_content="""
                {% bird test %}
                    {% bird:slot named_slot %}{% endbird:slot %}
                {% endbird %}
            """,
            template_context={
                "slots": {
                    "named_slot": "Content",
                },
            },
            expected="Content",
        )
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Outer slot replacement",
                component=TestComponent(
                    name="test",
                    content="""
                        {% bird:slot outer %}
                            Outer {% bird:slot inner %}Inner{% endbird:slot %} Content
                        {% endbird:slot %}
                    """,
                ),
                template_content="""
                    {% bird test %}
                        {% bird:slot outer %}Replaced Content{% endbird:slot %}
                    {% endbird %}
                """,
                expected="Replaced Content",
            ),
            TestComponentCase(
                description="Inner slot replacement",
                component=TestComponent(
                    name="test",
                    content="""
                        {% bird:slot outer %}
                            Outer {% bird:slot inner %}Inner{% endbird:slot %} Content
                        {% endbird:slot %}
                    """,
                ),
                template_content="""
                    {% bird test %}
                        {% bird:slot inner %}Replaced Content{% endbird:slot %}
                    {% endbird %}
                """,
                expected="Outer Replaced Content Content",
            ),
            TestComponentCase(
                description="Both slots replaced",
                component=TestComponent(
                    name="test",
                    content="""
                        {% bird:slot outer %}
                            Outer {% bird:slot inner %}Inner{% endbird:slot %} Content
                        {% endbird:slot %}
                    """,
                ),
                template_content="""
                    {% bird test %}
                        {% bird:slot outer %}
                            New {% bird:slot inner %}Nested{% endbird:slot %} Text
                        {% endbird:slot %}
                    {% endbird %}
                """,
                expected="New Nested Text",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_nested_slots(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Variable in slot",
                component=TestComponent(
                    name="test",
                    content="{% bird:slot %}{% endbird:slot %}",
                ),
                template_content="{% bird test %}{{ message }}{% endbird %}",
                template_context={"message": "Hello World"},
                expected="Hello World",
            ),
            TestComponentCase(
                description="Nested variable in slot",
                component=TestComponent(
                    name="test",
                    content="{% bird:slot %}{% endbird:slot %}",
                ),
                template_content="{% bird test %}{{ user.name }}{% endbird %}",
                template_context={"user": {"name": "John"}},
                expected="John",
            ),
            TestComponentCase(
                description="Template tag in slot",
                component=TestComponent(
                    name="test",
                    content="{% bird:slot %}{% endbird:slot %}",
                ),
                template_content="{% bird test %}{% if show %}Show{% endif %}{% endbird %}",
                template_context={"show": True},
                expected="Show",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_template_content(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    def test_too_many_args(self):
        with pytest.raises(TemplateSyntaxError):
            Template("{% bird:slot too many args %}{% endbird:slot %}")



from __future__ import annotations

from django.apps import apps

from django_bird.components import components

from .utils import TestComponent


def test_ready_scans_components(templates_dir):
    TestComponent(name="button", content="<button>Click me</button>").create(
        templates_dir
    )
    TestComponent(name="alert", content="<div>Alert</div>").create(templates_dir)

    components.reset()

    assert "button" not in components._components
    assert "alert" not in components._components

    apps.get_app_config("django_bird").ready()

    assert "button" in components._components
    assert "alert" in components._components



from __future__ import annotations

import multiprocessing
from collections.abc import Callable
from collections.abc import Generator
from collections.abc import Iterator
from itertools import chain
from multiprocessing import Pool
from pathlib import Path
from typing import Any
from typing import TypeGuard
from typing import final

from django.template.base import Node
from django.template.base import Template
from django.template.context import Context
from django.template.engine import Engine
from django.template.exceptions import TemplateDoesNotExist
from django.template.exceptions import TemplateSyntaxError
from django.template.loader_tags import ExtendsNode
from django.template.loader_tags import IncludeNode
from django.template.utils import get_app_template_dirs

from .conf import app_settings
from .templatetags.tags.bird import BirdNode
from .utils import get_files_from_dirs
from .utils import unique_ordered


def get_template_names(name: str) -> list[str]:
    """
    Generate a list of potential template names for a component.

    The function searches for templates in the following order (from most specific to most general):

    1. In a subdirectory named after the component, using the component name
    2. In the same subdirectory, using a fallback 'index.html'
    3. In parent directory for nested components
    4. In the base component directory, using the full component name

    The order of names is important as it determines the template resolution priority.
    This order allows for both direct matches and hierarchical component structures,
    with more specific paths taking precedence over more general ones.

    This order allows for:
    - Single file components
    - Multi-part components
    - Specific named files within component directories
    - Fallback default files for components

    For example:
    - For an "input" component, the ordering would be:
        1. `{component_dir}/input/input.html`
        2. `{component_dir}/input/index.html`
        3. `{component_dir}/input.html`
    - For an "input.label" component:
        1. `{component_dir}/input/label/label.html`
        2. `{component_dir}/input/label/index.html`
        3. `{component_dir}/input/label.html`
        4. `{component_dir}/input.label.html`

    Returns:
        list[str]: A list of potential template names in resolution order.
    """
    template_names: list[str] = []
    component_dirs = app_settings.get_component_directory_names()

    name_parts = name.split(".")
    path_name = "/".join(name_parts)

    for component_dir in component_dirs:
        potential_names = [
            f"{component_dir}/{path_name}/{name_parts[-1]}.html",
            f"{component_dir}/{path_name}/index.html",
            f"{component_dir}/{path_name}.html",
            f"{component_dir}/{name}.html",
        ]
        template_names.extend(potential_names)

    return unique_ordered(template_names)


def get_template_directories() -> Generator[Path, Any, None]:
    engine = Engine.get_default()
    for engine_dir in engine.dirs:
        yield Path(engine_dir)
    for app_dir in get_app_template_dirs("templates"):
        yield Path(app_dir)


def get_component_directories(
    template_dirs: Iterator[Path] | None = None,
) -> list[Path]:
    if template_dirs is None:
        template_dirs = get_template_directories()

    return [
        Path(template_dir) / component_dir
        for template_dir in template_dirs
        for component_dir in app_settings.get_component_directory_names()
    ]


def gather_bird_tag_template_usage() -> Generator[tuple[Path, set[str]], Any, None]:
    template_dirs = get_template_directories()
    templates = list(get_files_from_dirs(template_dirs))
    chunk_size = max(1, len(templates) // multiprocessing.cpu_count() * 2)
    chunks = [
        templates[i : i + chunk_size] for i in range(0, len(templates), chunk_size)
    ]
    with Pool() as pool:
        results = pool.map(_process_template_chunk, chunks)
    yield from chain.from_iterable(results)


def _process_template_chunk(  # pragma: no cover
    templates: list[tuple[Path, Path]],
) -> list[tuple[Path, set[str]]]:
    results: list[tuple[Path, set[str]]] = []
    for path, root in templates:
        visitor = NodeVisitor(Engine.get_default())
        template_name = str(path.relative_to(root))
        try:
            template = Engine.get_default().get_template(template_name)
        except (TemplateDoesNotExist, TemplateSyntaxError):
            continue
        context = Context()
        with context.bind_template(template):
            visitor.visit(template, context)
        if visitor.components:
            results.append((path, visitor.components))
    return results


NodeVisitorMethod = Callable[[Template | Node, Context], None]


def has_nodelist(node: Template | Node) -> TypeGuard[Template]:
    return hasattr(node, "nodelist")


@final
class NodeVisitor:  # pragma: no cover
    def __init__(self, engine: Engine):
        self.engine = engine
        self.components: set[str] = set()
        self.visited_templates: set[str] = set()

    def visit(self, node: Template | Node, context: Context) -> None:
        method_name = f"visit_{node.__class__.__name__}"
        visitor: NodeVisitorMethod = getattr(self, method_name, self.generic_visit)
        return visitor(node, context)

    def generic_visit(self, node: Template | Node, context: Context) -> None:
        if not has_nodelist(node) or node.nodelist is None:
            return
        for child_node in node.nodelist:
            self.visit(child_node, context)

    def visit_Template(self, template: Template, context: Context) -> None:
        if template.name is None or template.name in self.visited_templates:
            return
        self.visited_templates.add(template.name)
        self.generic_visit(template, context)

    def visit_BirdNode(self, node: BirdNode, context: Context) -> None:
        component_name = node.name.strip("\"'")
        self.components.add(component_name)
        self.generic_visit(node, context)

    def visit_ExtendsNode(self, node: ExtendsNode, context: Context) -> None:
        parent_template = node.get_parent(context)
        self.visit(parent_template, context)
        self.generic_visit(node, context)

    def visit_IncludeNode(self, node: IncludeNode, context: Context) -> None:
        try:
            included_templates = node.template.resolve(context)
            if not isinstance(included_templates, list | tuple):
                included_templates = [included_templates]
            for template_name in included_templates:
                included_template = self.engine.get_template(template_name)
                self.visit(included_template, context)
        except Exception:
            pass
        self.generic_visit(node, context)



# pyright: reportAny=false
from __future__ import annotations

from typing import final

from django import template
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context

from django_bird._typing import override
from django_bird.components import components
from django_bird.staticfiles import AssetType

CSS_TAG = "bird:css"
JS_TAG = "bird:js"


def do_asset(_parser: Parser, token: Token) -> AssetNode:
    bits = token.split_contents()
    if len(bits) < 1:
        msg = "bird:assets tag requires at least one argument"
        raise template.TemplateSyntaxError(msg)
    tag_name = bits[0]
    asset_type = AssetType.from_tag_name(tag_name)
    return AssetNode(asset_type)


@final
class AssetNode(template.Node):
    def __init__(self, asset_type: AssetType):
        self.asset_type = asset_type

    @override
    def render(self, context: Context) -> str:
        template = getattr(context, "template", None)
        if not template:
            return ""
        used_components = components.get_component_usage(template.origin.name)
        assets = set(
            asset
            for component in used_components
            for asset in component.assets
            if asset.type == self.asset_type
        )
        if not assets:
            return ""
        rendered = [asset.render() for asset in sorted(assets, key=lambda a: a.path)]
        return "\n".join(rendered)



from __future__ import annotations

from django.template.loader import get_template

from django_bird.components import components


def test_template_inheritance_assets(example_template):
    components.discover_components()

    rendered = get_template(example_template.template.name).render({})

    assert all(
        asset.url in rendered
        for component in example_template.used_components
        for asset in component.assets
    )
    assert not any(
        asset.url in rendered
        for component in example_template.unused_components
        for asset in component.assets
    )



# pyright: reportAny=false
from __future__ import annotations

from typing import final

from django import template
from django.template.base import NodeList
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context

from django_bird._typing import TagBits
from django_bird._typing import override

TAG = "bird"
END_TAG = "endbird"


def do_bird(parser: Parser, token: Token) -> BirdNode:
    _tag, *bits = token.split_contents()
    if not bits:
        msg = f"{TAG} tag requires at least one argument"
        raise template.TemplateSyntaxError(msg)

    name = bits.pop(0)
    attrs: TagBits = []
    isolated_context = False

    for bit in bits:
        match bit:
            case "only":
                isolated_context = True
            case "/":
                continue
            case _:
                attrs.append(bit)

    nodelist = parse_nodelist(bits, parser)
    return BirdNode(name, attrs, nodelist, isolated_context)


def parse_nodelist(bits: TagBits, parser: Parser) -> NodeList | None:
    # self-closing tag
    # {% bird name / %}
    if len(bits) > 0 and bits[-1] == "/":
        nodelist = None
    else:
        nodelist = parser.parse((END_TAG,))
        parser.delete_first_token()
    return nodelist


@final
class BirdNode(template.Node):
    def __init__(
        self,
        name: str,
        attrs: TagBits,
        nodelist: NodeList | None,
        isolated_context: bool = False,
    ) -> None:
        self.name = name
        self.attrs = attrs
        self.nodelist = nodelist
        self.isolated_context = isolated_context

    @override
    def render(self, context: Context) -> str:
        from django_bird.components import components

        component_name = self.get_component_name(context)
        component = components.get_component(component_name)
        bound_component = component.get_bound_component(node=self)

        if self.isolated_context:
            return bound_component.render(context.new())
        else:
            return bound_component.render(context)

    def get_component_name(self, context: Context) -> str:
        try:
            name = template.Variable(self.name).resolve(context)
        except template.VariableDoesNotExist:
            name = self.name
        return name



from __future__ import annotations

import re
from typing import Any
from typing import final

from django import template
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context

from django_bird._typing import override

register = template.Library()

TAG = "bird:var"
END_TAG = "endbird:var"

OPERATOR_PATTERN = re.compile(r"(\w+)\s*(\+=|=)\s*(.+)")


def do_var(parser: Parser, token: Token):
    _tag, *bits = token.split_contents()
    if not bits:
        msg = f"'{TAG}' tag requires an assignment"
        raise template.TemplateSyntaxError(msg)

    var_assignment = bits.pop(0)
    match = re.match(OPERATOR_PATTERN, var_assignment)
    if not match:
        msg = (
            f"Invalid assignment in '{TAG}' tag: {var_assignment}. "
            f"Expected format: {TAG} variable='value' or {TAG} variable+='value'."
        )
        raise template.TemplateSyntaxError(msg)

    var_name, operator, var_value = match.groups()
    var_value = var_value.strip()
    value = parser.compile_filter(var_value)

    return VarNode(var_name, operator, value)


def do_end_var(_parser: Parser, token: Token):
    _tag, *bits = token.split_contents()
    if not bits:
        msg = f"{token.contents.split()[0]} tag requires a variable name"
        raise template.TemplateSyntaxError(msg)

    var_name = bits.pop(0)

    return EndVarNode(var_name)


@final
class VarNode(template.Node):
    def __init__(self, name: str, operator: str, value: Any):
        self.name = name
        self.operator = operator
        self.value = value

    @override
    def render(self, context: Context) -> str:
        if "vars" not in context:
            context["vars"] = {}

        value = self.value.resolve(context)

        if self.operator == "+=":
            previous = context["vars"].get(self.name, "")
            value = f"{previous}{value}"

        context["vars"][self.name] = value
        return ""


@final
class EndVarNode(template.Node):
    def __init__(self, name: str):
        self.name = name

    @override
    def render(self, context: Context) -> str:
        if "vars" in context and self.name in context["vars"]:
            del context["vars"][self.name]
        return ""



from __future__ import annotations

import importlib

import pluggy

from django_bird.plugins import hookspecs

pm = pluggy.PluginManager("django_bird")
pm.add_hookspecs(hookspecs)

pm.load_setuptools_entrypoints("django_bird")

DEFAULT_PLUGINS: list[str] = [
    "django_bird.staticfiles",
]

for plugin in DEFAULT_PLUGINS:
    mod = importlib.import_module(plugin)
    pm.register(mod, plugin)



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from typing import Any

from django import template
from django.template.context import Context
from django.utils.safestring import SafeString
from django.utils.safestring import mark_safe

from .templatetags.tags.bird import BirdNode
from .templatetags.tags.prop import PropNode

if TYPE_CHECKING:
    from django_bird.components import Component


@dataclass
class Params:
    attrs: list[Param] = field(default_factory=list)
    props: list[Param] = field(default_factory=list)

    def render_props(self, component: Component, context: Context):
        if component.nodelist is None:
            return

        attrs_to_remove = set()

        for node in component.nodelist:
            if not isinstance(node, PropNode):
                continue

            value = Value(node.default)

            for idx, attr in enumerate(self.attrs):
                if node.name == attr.name:
                    resolved = attr.value.resolve(context)
                    if resolved is not None:
                        value = attr.value
                    attrs_to_remove.add(idx)

            self.props.append(Param(name=node.name, value=value))

        for idx in sorted(attrs_to_remove, reverse=True):
            self.attrs.pop(idx)

        return {prop.name: prop.render_prop(context) for prop in self.props}

    def render_attrs(self, context: Context) -> SafeString:
        rendered = " ".join(attr.render_attr(context) for attr in self.attrs)
        return mark_safe(rendered)

    @classmethod
    def from_node(cls, node: BirdNode) -> Params:
        return cls(
            attrs=[Param.from_bit(bit) for bit in node.attrs],
            props=[],
        )


@dataclass
class Param:
    name: str
    value: Value

    def render_attr(self, context: Context) -> str:
        value = self.value.resolve(context)
        if value is None:
            return ""
        name = self.name.replace("_", "-")
        if value is True:
            return name
        return f'{name}="{value}"'

    def render_prop(self, context: Context) -> str | bool | None:
        return self.value.resolve(context)

    @classmethod
    def from_bit(cls, bit: str) -> Param:
        if "=" in bit:
            name, raw_value = bit.split("=", 1)
            value = Value(raw_value.strip())
        else:
            name, value = bit, Value(True)
        return cls(name, value)


@dataclass
class Value:
    raw: str | bool | None

    def resolve(self, context: Context | dict[str, Any]) -> Any:
        match (self.raw, self.is_quoted):
            case (None, _):
                return None

            case (str(raw_str), False) if raw_str == "False":
                return None
            case (str(raw_str), False) if raw_str == "True":
                return True

            case (bool(b), _):
                return b if b else None

            case (str(raw_str), False):
                try:
                    return template.Variable(raw_str).resolve(context)
                except template.VariableDoesNotExist:
                    return raw_str

            case (_, True):
                return str(self.raw)[1:-1]

    @property
    def is_quoted(self) -> bool:
        if self.raw is None or isinstance(self.raw, bool):
            return False

        return self.raw.startswith(("'", '"')) and self.raw.endswith(self.raw[0])



from __future__ import annotations

from contextlib import suppress
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any
from typing import final

import django.template
from django.conf import settings

from ._typing import override
from .utils import unique_ordered

DJANGO_BIRD_SETTINGS_NAME = "DJANGO_BIRD"

DJANGO_BIRD_BUILTINS = "django_bird.templatetags.django_bird"
DJANGO_BIRD_FINDER = "django_bird.staticfiles.BirdAssetFinder"


@dataclass
class AppSettings:
    COMPONENT_DIRS: list[Path | str] = field(default_factory=list)
    ENABLE_AUTO_CONFIG: bool = True
    ENABLE_BIRD_ATTRS: bool = True
    _configurator: AutoConfigurator = field(init=False)

    def __post_init__(self):
        self._configurator = AutoConfigurator(self)

    @override
    def __getattribute__(self, __name: str) -> object:
        user_settings = getattr(settings, DJANGO_BIRD_SETTINGS_NAME, {})
        return user_settings.get(__name, super().__getattribute__(__name))

    def autoconfigure(self) -> None:
        if not self.ENABLE_AUTO_CONFIG:
            return

        self._configurator.autoconfigure()

    def get_component_directory_names(self):
        return unique_ordered([*self.COMPONENT_DIRS, "bird"])


@final
class AutoConfigurator:
    def __init__(self, app_settings: AppSettings) -> None:
        self.app_settings = app_settings
        self._configured = False

    def autoconfigure(self) -> None:
        self.configure_templates()
        self.configure_staticfiles()
        self._configured = True

    def configure_templates(self) -> None:
        template_config = None

        for config in settings.TEMPLATES:
            engine_name = config.get("NAME") or config["BACKEND"].split(".")[-2]
            if engine_name == "django":
                template_config = config
                break

        if template_config is None:
            return

        options = template_config.setdefault("OPTIONS", {})

        self.configure_builtins(options)

        # Force re-evaluation of settings.TEMPLATES because EngineHandler caches it.
        with suppress(AttributeError):  # pragma: no cover
            del django.template.engines.templates
            django.template.engines._engines = {}  # type: ignore[attr-defined]

    def configure_builtins(self, options: dict[str, Any]) -> None:
        builtins = options.setdefault("builtins", [])

        builtins_already_configured = DJANGO_BIRD_BUILTINS in builtins

        if not builtins_already_configured:
            builtins.append(DJANGO_BIRD_BUILTINS)

    def configure_staticfiles(self) -> None:
        finders_already_configured = DJANGO_BIRD_FINDER in settings.STATICFILES_FINDERS

        if not finders_already_configured:
            settings.STATICFILES_FINDERS.append(DJANGO_BIRD_FINDER)


app_settings = AppSettings()



from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path

from pluggy import HookspecMarker

from django_bird.staticfiles import Asset

hookspec = HookspecMarker("django_bird")


@hookspec
def collect_component_assets(template_path: Path) -> Iterable[Asset]:
    """Collect all assets associated with a component.

    This hook is called for each component template to gather its associated static assets.
    Implementations should scan for and return any CSS, JavaScript or other static files
    that belong to the component and return a list/set/other iterable of `Asset` objects.
    """



from __future__ import annotations

import pytest
from django.template import Context
from django.template import Library
from django.template import Template
from django.template.base import Parser
from django.template.base import Token
from django.template.base import TokenType
from django.template.engine import Engine
from django.template.exceptions import TemplateDoesNotExist
from django.template.exceptions import TemplateSyntaxError

from django_bird.components import Component
from django_bird.templatetags.tags.bird import END_TAG
from django_bird.templatetags.tags.bird import TAG
from django_bird.templatetags.tags.bird import BirdNode
from django_bird.templatetags.tags.bird import do_bird
from tests.utils import TestComponent
from tests.utils import TestComponentCase
from tests.utils import normalize_whitespace


class TestTagParsing:
    @pytest.mark.parametrize(
        "name,expected",
        [
            ("button", "button"),
            ("'button'", "'button'"),
            ('"button"', '"button"'),
            ("button.label", "button.label"),
        ],
    )
    def test_name_do_bird(self, name, expected):
        start_token = Token(TokenType.BLOCK, f"{TAG} {name}")
        end_token = Token(TokenType.BLOCK, f"{END_TAG} {name}")

        parser = Parser([end_token])

        node = do_bird(parser, start_token)

        assert node.name == expected

    def test_missing_name_do_bird(self):
        start_token = Token(TokenType.BLOCK, TAG)
        end_token = Token(TokenType.BLOCK, END_TAG)

        parser = Parser([end_token])

        with pytest.raises(TemplateSyntaxError):
            do_bird(parser, start_token)

    @pytest.mark.parametrize(
        "params,expected_attrs",
        [
            ('class="btn"', ['class="btn"']),
            ("class='btn'", ["class='btn'"]),
            ('class="btn" id="my-btn"', ['class="btn"', 'id="my-btn"']),
            ("disabled", ["disabled"]),
            ("class=dynamic", ["class=dynamic"]),
            ("class=item.name id=user.id", ["class=item.name", "id=user.id"]),
        ],
    )
    def test_attrs_do_bird(self, params, expected_attrs):
        start_token = Token(TokenType.BLOCK, f"{TAG} button {params}")
        end_token = Token(TokenType.BLOCK, f"{END_TAG} button")

        parser = Parser([end_token])

        node = do_bird(parser, start_token)

        assert node.attrs == expected_attrs

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Basic component name",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button %}
                        Click me
                    {% endbird %}
                """,
                expected="<button>Click me</button>",
            ),
            TestComponentCase(
                description="Single-quoted component name",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird 'button' %}
                        Click me
                    {% endbird %}
                """,
                expected="<button>Click me</button>",
            ),
            TestComponentCase(
                description="Double-quoted component name",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird "button" %}
                        Click me
                    {% endbird %}
                """,
                expected="<button>Click me</button>",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_basic_name_templatetag(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(
            test_case.template_content,
        )
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    def test_nested_name_templatetag(self, templates_dir):
        test_case = TestComponentCase(
            description="Nested component name",
            component=TestComponent(
                name="label",
                content="""
                    <span>
                        {{ slot }}
                    </span>
                """,
                sub_dir="button",
            ),
            template_content="""
                {% bird button.label %}
                    Click me
                {% endbird %}
            """,
            expected="<span>Click me</span>",
        )
        test_case.component.create(templates_dir)

        template = Template(
            test_case.template_content,
        )
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Dynamic component name using with",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% with dynamic_name='button' %}
                        {% bird dynamic_name %}
                            Click me
                        {% endbird %}
                    {% endwith %}
                """,
                expected="<button>Click me</button>",
            ),
            TestComponentCase(
                description="Dynamic component name from context",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird dynamic-name %}
                        Click me
                    {% endbird %}
                """,
                template_context={"dynamic-name": "button"},
                expected="<button>Click me</button>",
            ),
            TestComponentCase(
                description="Quoted component name should not be dynamic",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird "button" %}
                        Click me
                    {% endbird %}
                """,
                template_context={"button": "dynamic_name"},
                expected="<button>Click me</button>",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_dynamic_name_template_context_templatetag(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(
            test_case.template_content,
        )
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    def test_dynamic_name_with_string(self, templates_dir):
        button = TestComponent(
            name="button",
            content="""
                <button>
                    {{ slot }}
                </button>
            """,
        )
        not_button = TestComponent(
            name="not_button",
            content="""
                <div>
                    {{ slot }}
                </div>
            """,
        )
        button.create(templates_dir)
        not_button.create(templates_dir)

        template = Template("{% bird 'button' %}Click me{% endbird %}")
        rendered = template.render(Context({"button": "not_button"}))

        assert normalize_whitespace(rendered) == "<button>Click me</button>"
        assert normalize_whitespace(rendered) != "<div>Click me</div>"

    def test_nonexistent_name_templatetag(self, templates_dir):
        template = Template("{% bird nonexistent %}Content{% endbird %}")

        with pytest.raises(TemplateDoesNotExist):
            template.render(Context({}))


def test_self_closing_tag(templates_dir):
    test_case = TestComponentCase(
        component=TestComponent(
            name="image",
            content="<img src='image' />",
        ),
        template_content="{% bird image src='image' / %}",
        expected="<img src='image' />",
    )
    test_case.component.create(templates_dir)

    template = Template(
        test_case.template_content,
    )
    rendered = template.render(Context(test_case.template_context))

    assert normalize_whitespace(rendered) == test_case.expected


class TestAttributes:
    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Renders class attribute",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class='btn' %}
                        Click me
                    {% endbird %}
                """,
                expected='<button class="btn">Click me</button>',
            ),
            TestComponentCase(
                description="Renders multiple attributes",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class='btn' id='my-btn' %}
                        Click me
                    {% endbird %}
                """,
                expected='<button class="btn" id="my-btn">Click me</button>',
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_basic(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Implicit boolean attribute",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button disabled %}
                        Click me
                    {% endbird %}
                """,
                expected="<button disabled>Click me</button>",
            ),
            TestComponentCase(
                description="Explicit boolean True",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button disabled=True %}
                        Click me
                    {% endbird %}
                """,
                expected="<button disabled>Click me</button>",
            ),
            TestComponentCase(
                description="Omits boolean False",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button disabled=False %}
                        Click me
                    {% endbird %}
                """,
                expected="<button>Click me</button>",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_boolean(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Dynamic attrs using with",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% with dynamic_class='btn-primary' %}
                        {% bird button class=dynamic_class %}
                            Click me
                        {% endbird %}
                    {% endwith %}
                """,
                expected='<button class="btn-primary">Click me</button>',
            ),
            TestComponentCase(
                description="Dynamic attrs using template context",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class=dynamic_class %}
                        Click me
                    {% endbird %}
                """,
                template_context={"dynamic_class": "btn-primary"},
                expected='<button class="btn-primary">Click me</button>',
            ),
            TestComponentCase(
                description="Literal string over context value",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class='dynamic_class' %}
                        Click me
                    {% endbird %}
                """,
                template_context={"dynamic_class": "btn-primary"},
                expected='<button class="dynamic_class">Click me</button>',
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_dynamic_template_context(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Resolves nested object attribute",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class=btn.class %}
                        Click me
                    {% endbird %}
                """,
                template_context={
                    "btn": {
                        "class": "btn-success",
                    },
                },
                expected='<button class="btn-success">Click me</button>',
            ),
            TestComponentCase(
                description="Preserves literal nested path",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class='btn.class' %}
                        Click me
                    {% endbird %}
                """,
                template_context={
                    "btn": {
                        "class": "btn-success",
                    },
                },
                expected='<button class="btn.class">Click me</button>',
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_nested_template_context(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Handles undefined variable",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class=undefined_class %}
                        Click me
                    {% endbird %}
                """,
                expected='<button class="undefined_class">Click me</button>',
            ),
            TestComponentCase(
                description="Handles missing nested attribute",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class=missing.attr %}
                        Click me
                    {% endbird %}
                """,
                expected='<button class="missing.attr">Click me</button>',
            ),
            TestComponentCase(
                description="Handles empty nested object",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            Click me
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class=user.preferences.theme %}
                        Click me
                    {% endbird %}
                """,
                template_context={
                    "user": {
                        "preferences": {},
                    }
                },
                expected='<button class="user.preferences.theme">Click me</button>',
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_error_handling(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "attr_app_setting,expected",
        [
            ({"ENABLE_BIRD_ATTRS": True}, True),
            ({"ENABLE_BIRD_ATTRS": False}, False),
        ],
    )
    def test_data_bird_attributes(
        self,
        attr_app_setting,
        expected,
        override_app_settings,
        templates_dir,
    ):
        button = TestComponent(
            name="button",
            content="""
                <button {{ attrs }}>
                    {{ slot }}
                </button>
            """,
        ).create(templates_dir)

        template = Template("{% bird 'button' %}Click me{% endbird %}")

        with override_app_settings(**attr_app_setting):
            rendered = template.render(Context({}))

        comp = Component.from_name(button.name)

        assert (f'data-bird-id="{comp.id}' in rendered) is expected
        assert (f"data-bird-{comp.data_attribute_name}" in rendered) is expected


class TestProperties:
    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Basic prop passing",
                component=TestComponent(
                    name="button",
                    content="""
                        {% bird:prop id %}
                        <button id='{{ props.id }}'>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button id='foo' %}
                        Click me
                    {% endbird %}
                """,
                expected="<button id='foo'>Click me</button>",
            ),
            TestComponentCase(
                description="Default prop value",
                component=TestComponent(
                    name="button",
                    content="""
                        {% bird:prop id='default' %}
                        <button id='{{ props.id }}'>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button %}
                        Click me
                    {% endbird %}
                """,
                expected="<button id='default'>Click me</button>",
            ),
            TestComponentCase(
                description="Override default prop",
                component=TestComponent(
                    name="button",
                    content="""
                        {% bird:prop id='default' %}
                        <button id='{{ props.id }}'>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button id='foo' %}
                        Click me
                    {% endbird %}
                """,
                expected="<button id='foo'>Click me</button>",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_basic(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Dynamic props using with",
                component=TestComponent(
                    name="button",
                    content="""
                        {% bird:prop class %}
                        <button class='{{ props.class }}'>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% with dynamic_class='btn-primary' %}
                        {% bird button class=dynamic_class %}
                            Click me
                        {% endbird %}
                    {% endwith %}
                """,
                expected="<button class='btn-primary'>Click me</button>",
            ),
            TestComponentCase(
                description="Dynamic props using template context",
                component=TestComponent(
                    name="button",
                    content="""
                        {% bird:prop class %}
                        <button class='{{ props.class }}'>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class=dynamic_class %}
                        Click me
                    {% endbird %}
                """,
                template_context={
                    "dynamic_class": "btn-primary",
                },
                expected="<button class='btn-primary'>Click me</button>",
            ),
            TestComponentCase(
                description="Literal string over context value",
                component=TestComponent(
                    name="button",
                    content="""
                        {% bird:prop class %}
                        <button class='{{ props.class }}'>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class='dynamic_class' %}
                        Click me
                    {% endbird %}
                """,
                template_context={
                    "dynamic_class": "btn-primary",
                },
                expected="<button class='dynamic_class'>Click me</button>",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_dynamic(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Undefined variable",
                component=TestComponent(
                    name="button",
                    content="""
                        {% bird:prop class %}
                        <button class='{{ props.class }}'>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class=undefined_class %}
                        Click me
                    {% endbird %}
                """,
                expected="<button class='undefined_class'>Click me</button>",
            ),
            TestComponentCase(
                description="Missing nested attribute",
                component=TestComponent(
                    name="button",
                    content="""
                        {% bird:prop class %}
                        <button class='{{ props.class }}'>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class=missing.attr %}
                        Click me
                    {% endbird %}
                """,
                expected="<button class='missing.attr'>Click me</button>",
            ),
            TestComponentCase(
                description="Empty nested object",
                component=TestComponent(
                    name="button",
                    content="""
                        {% bird:prop class %}
                        <button class='{{ props.class }}'>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class=user.preferences.theme %}
                        Click me
                    {% endbird %}
                """,
                template_context={"user": {"preferences": {}}},
                expected="<button class='user.preferences.theme'>Click me</button>",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_error_handling(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected


def test_attrs_and_props(templates_dir):
    test_case = TestComponentCase(
        component=TestComponent(
            name="button",
            content="""
                    {% bird:prop id="default" %}
                    {% bird:prop class="btn" %}

                    <button id="{{ props.id }}" class="{{ props.class }}" {{ attrs }}>
                        {{ slot }}
                    </button>
                """,
        ),
        template_content="""
            {% bird button data-test='value' %}
                Click me
            {% endbird %}
        """,
        expected='<button id="default" class="btn" data-test="value">Click me</button>',
    )
    test_case.component.create(templates_dir)

    template = Template(test_case.template_content)
    rendered = template.render(Context(test_case.template_context))

    assert normalize_whitespace(rendered) == test_case.expected


class TestSlots:
    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Default slot renders content",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button %}
                        Click me
                    {% endbird %}
                """,
                expected="<button>Click me</button>",
            ),
            TestComponentCase(
                description="Empty named slot declaration",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            {% bird:slot %}{% endbird:slot %}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button %}
                        Click me
                    {% endbird %}
                """,
                expected="<button>Click me</button>",
            ),
            TestComponentCase(
                description="Named default slot",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            {% bird:slot default %}{% endbird:slot %}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button %}
                        Click me
                    {% endbird %}
                """,
                expected="<button>Click me</button>",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_default(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Renders content in named slot alongside default slot",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            <span>
                                {% bird:slot leading-icon %}{% endbird:slot %}
                            </span>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button %}
                        {% bird:slot leading-icon %}
                            Icon here
                        {% endbird:slot %}
                        Click me
                    {% endbird %}
                """,
                expected="<button><span>Icon here</span>Click me</button>",
            ),
            TestComponentCase(
                description="Renders multiple named slots in different positions",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            {% bird:slot prefix %}{% endbird:slot %}
                            <span>{{ slot }}</span>
                            {% bird:slot suffix %}{% endbird:slot %}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button %}
                        {% bird:slot prefix %}Before{% endbird:slot %}
                        Click me
                        {% bird:slot suffix %}After{% endbird:slot %}
                    {% endbird %}
                """,
                expected="<button>Before<span>Click me</span>After</button>",
            ),
            TestComponentCase(
                description="Renders named slots with component variables",
                component=TestComponent(
                    name="button",
                    content="""
                        <button {{ attrs }}>
                            {% bird:slot prefix %}{% endbird:slot %}
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button class="primary" %}
                        {% bird:slot prefix %}→{% endbird:slot %}
                        Submit
                    {% endbird %}
                """,
                expected='<button class="primary">→ Submit</button>',
            ),
            TestComponentCase(
                description="Handles empty named slots by rendering nothing",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            {% bird:slot prefix %}{% endbird:slot %}
                            {{ slot }}
                            {% bird:slot suffix %}{% endbird:slot %}
                        </button>
                    """,
                ),
                template_content="""
                    {% bird button %}
                        Click me
                    {% endbird %}
                """,
                expected="<button>Click me</button>",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_named(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.fixture
    def fancy_filter(self):
        register = Library()

        @register.filter
        def make_fancy(value):
            return f"✨{value}✨"

        engine = Engine.get_default()
        engine.template_libraries["custom_filters"] = register

        yield

        del engine.template_libraries["custom_filters"]

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Filter in default slot",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% load custom_filters %}
                    {% bird button %}
                        {{ text|make_fancy }}
                    {% endbird %}
                """,
                template_context={"text": "click me"},
                expected="<button>✨click me✨</button>",
            ),
            TestComponentCase(
                description="Filter in named slot",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            {% bird:slot label %}{% endbird:slot %}
                        </button>
                    """,
                ),
                template_content="""
                    {% load custom_filters %}
                    {% bird button %}
                        {% bird:slot label %}
                            {{ text|make_fancy }}
                        {% endbird:slot %}
                    {% endbird %}
                """,
                template_context={"text": "click me"},
                expected="<button>✨click me✨</button>",
            ),
            TestComponentCase(
                description="Filters in multiple slots",
                component=TestComponent(
                    name="button",
                    content="""
                        <button>
                            <span>
                                {% bird:slot icon %}{% endbird:slot %}
                            </span>
                            {{ slot }}
                        </button>
                    """,
                ),
                template_content="""
                    {% load custom_filters %}
                    {% bird button %}
                        {% bird:slot icon %}
                            {{ icon_text|make_fancy }}
                        {% endbird:slot %}
                        {{ text|make_fancy }}
                    {% endbird %}
                """,
                template_context={"text": "click me", "icon_text": "icon"},
                expected="<button><span>✨icon✨</span>✨click me✨</button>",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_with_outside_templatetag(self, test_case, fancy_filter, templates_dir):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected


def test_nested_components_with_loops(templates_dir):
    nav = TestComponent(
        name="nav",
        content="""
            <nav>
                {{ slot }}
            </nav>
        """,
    )
    nav.create(templates_dir)

    nav_item = TestComponent(
        name="item",
        content="""
            {% bird:prop href="#" %}

            <a href="{{ props.href }}">{{ slot }}</a>
        """,
        sub_dir="nav",
    )
    nav_item.create(templates_dir)

    template = Template("""
        {% bird nav %}
            {% for item in items %}
                {% bird nav.item href=item.url %}
                    {{ item.title }}
                {% endbird nav.item %}
            {% endfor %}
        {% endbird nav %}
    """)

    context = Context(
        {
            "items": [
                {"url": "/", "title": "Home"},
                {"url": "/admin", "title": "Admin"},
            ]
        }
    )

    rendered = template.render(context)

    assert normalize_whitespace(rendered) == normalize_whitespace("""
        <nav>
            <a href="/">Home</a>
            <a href="/admin">Admin</a>
        </nav>
    """)


@pytest.mark.parametrize(
    "test_case",
    [
        TestComponentCase(
            description="Access parent context variable",
            component=TestComponent(
                name="button",
                content="""
                    <button>
                        {{ user.name }}
                    </button>
                """,
            ),
            template_content="""
                {% bird button %}{% endbird %}
            """,
            template_context={"user": {"name": "John"}},
            expected="<button>John</button>",
        ),
        TestComponentCase(
            description="Access parent context in slot",
            component=TestComponent(
                name="button",
                content="""
                    <button>
                        {{ slot }}
                    </button>
                """,
            ),
            template_content="""
                {% bird button %}
                    {{ user.name }}
                {% endbird %}
            """,
            template_context={"user": {"name": "John"}},
            expected="<button>John</button>",
        ),
        TestComponentCase(
            description="Access parent context in named slot",
            component=TestComponent(
                name="button",
                content="""
                    <button>
                        {% bird:slot prefix %}{% endbird:slot %}
                        {{ slot }}
                    </button>
                """,
            ),
            template_content="""
                {% bird button %}
                    {% bird:slot prefix %}{{ user.role }}{% endbird:slot %}
                    {{ user.name }}
                {% endbird %}
            """,
            template_context={"user": {"name": "John", "role": "Admin"}},
            expected="<button>Admin John</button>",
        ),
        TestComponentCase(
            description="Component-specific context overrides parent context values",
            component=TestComponent(
                name="button",
                content="""
                    <button {{ attrs }}>
                        {{ slot }}
                    </button>
                """,
            ),
            template_content="""
                {% bird button id="foo" %}
                    Component Content
                {% endbird %}
            """,
            template_context={
                "slot": "Parent Content",
                "attrs": 'id="bar"',
            },
            expected='<button id="foo">Component Content</button>',
        ),
    ],
    ids=lambda x: x.description,
)
def test_parent_context_access(test_case, templates_dir):
    test_case.component.create(templates_dir)

    template = Template(test_case.template_content)
    rendered = template.render(Context(test_case.template_context))

    assert normalize_whitespace(rendered) == test_case.expected


@pytest.mark.parametrize(
    "test_case",
    [
        TestComponentCase(
            description="Only flag prevents access to parent context",
            component=TestComponent(
                name="button",
                content="""
                    <button>
                        {{ user.name|default:"Anonymous" }}
                    </button>
                """,
            ),
            template_content="""
                {% bird button only %}{% endbird %}
            """,
            template_context={"user": {"name": "John"}},
            expected="<button>Anonymous</button>",
        ),
        TestComponentCase(
            description="Only flag still allows props and slots",
            component=TestComponent(
                name="button",
                content="""
                    {% bird:prop label %}
                    <button {{ attrs }}>
                        {{ props.label }}
                        {{ slot }}
                        {{ user.name|default:"Anonymous" }}
                    </button>
                """,
            ),
            template_content="""
                {% bird button id="foo" label="Click" only %}
                    Content
                {% endbird %}
            """,
            template_context={
                "props": {
                    "label": "Outside",
                },
                "user": {"name": "John"},
            },
            expected='<button id="foo">Click Content Anonymous</button>',
        ),
        TestComponentCase(
            description="Only flag with named slots",
            component=TestComponent(
                name="button",
                content="""
                    <button>
                        {% bird:slot prefix %}{% endbird:slot %}
                        {{ user.name|default:"Anonymous" }}
                    </button>
                """,
            ),
            template_content="""
                {% bird button only %}
                    {% bird:slot prefix %}{{ user.role|default:"User" }}{% endbird:slot %}
                {% endbird %}
            """,
            template_context={"user": {"name": "John", "role": "Admin"}},
            expected="<button>User Anonymous</button>",
        ),
        TestComponentCase(
            description="Only flag with self-closing tag",
            component=TestComponent(
                name="button",
                content="""
                    <button>
                        {{ user.name|default:"Anonymous" }}
                    </button>
                """,
            ),
            template_content="""
                {% bird button only / %}
            """,
            template_context={"user": {"name": "John"}},
            expected="<button>Anonymous</button>",
        ),
    ],
    ids=lambda x: x.description,
)
def test_only_flag(test_case, templates_dir):
    test_case.component.create(templates_dir)

    template = Template(test_case.template_content)
    rendered = template.render(Context(test_case.template_context))

    assert normalize_whitespace(rendered) == test_case.expected


class TestBirdNode:
    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Static name",
                component=TestComponent(
                    name="button",
                    content="",
                ),
                template_content="",
                expected="button",
            ),
            TestComponentCase(
                description="Dynamic name",
                component=TestComponent(
                    name="button",
                    content="",
                ),
                template_content="",
                template_context={
                    "dynamic-name": "button",
                },
                expected="button",
            ),
        ],
    )
    def test_get_component_name(self, test_case, templates_dir):
        test_case.component.create(templates_dir)

        node = BirdNode(name=test_case.component.name, attrs=[], nodelist=None)

        component_name = node.get_component_name(
            context=Context(test_case.template_context)
        )

        assert component_name == test_case.expected



from __future__ import annotations

import json
import os
from pathlib import Path

import nox

nox.options.default_venv_backend = "uv|virtualenv"
nox.options.reuse_existing_virtualenvs = True

PY310 = "3.10"
PY311 = "3.11"
PY312 = "3.12"
PY313 = "3.13"
PY_VERSIONS = [PY310, PY311, PY312, PY313]
PY_DEFAULT = PY_VERSIONS[0]
PY_LATEST = PY_VERSIONS[-1]

DJ42 = "4.2"
DJ50 = "5.0"
DJ51 = "5.1"
DJ52 = "5.2a1"
DJMAIN = "main"
DJMAIN_MIN_PY = PY312
DJ_VERSIONS = [DJ42, DJ50, DJ51, DJ52, DJMAIN]
DJ_LTS = [
    version for version in DJ_VERSIONS if version.endswith(".2") and version != DJMAIN
]
DJ_DEFAULT = DJ_LTS[0]
DJ_LATEST = DJ_VERSIONS[-2]


def version(ver: str) -> tuple[int, ...]:
    """Convert a string version to a tuple of ints, e.g. "3.10" -> (3, 10)"""
    return tuple(map(int, ver.split(".")))


def should_skip(python: str, django: str) -> bool:
    """Return True if the test should be skipped"""

    if django == DJMAIN and version(python) < version(DJMAIN_MIN_PY):
        # Django main requires Python 3.10+
        return True

    if django == DJ52 and version(python) < version(PY310):
        # Django 5.2a1 requires Python 3.10+
        return True

    if django == DJ51 and version(python) < version(PY310):
        # Django 5.1 requires Python 3.10+
        return True

    if django == DJ50 and version(python) < version(PY310):
        # Django 5.0 requires Python 3.10+
        return True

    return False


@nox.session
def test(session):
    session.notify(f"tests(python='{PY_DEFAULT}', django='{DJ_DEFAULT}')")


@nox.session
@nox.parametrize(
    "python,django",
    [
        (python, django)
        for python in PY_VERSIONS
        for django in DJ_VERSIONS
        if not should_skip(python, django)
    ],
)
def tests(session, django):
    session.run_install(
        "uv",
        "sync",
        "--frozen",
        "--inexact",
        "--no-install-package",
        "django",
        "--python",
        session.python,
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )

    if django == DJMAIN:
        session.install(
            "django @ https://github.com/django/django/archive/refs/heads/main.zip"
        )
    else:
        session.install(f"django=={django}")

    command = ["python", "-m", "pytest"]
    if session.posargs:
        args = []
        for arg in session.posargs:
            if arg:
                args.extend(arg.split(" "))
        command.extend(args)
    session.run(*command)


@nox.session
def coverage(session):
    session.run_install(
        "uv",
        "sync",
        "--frozen",
        "--python",
        PY_DEFAULT,
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )

    try:
        command = ["python", "-m", "pytest", "--cov", "--cov-report="]
        if session.posargs:
            args = []
            for arg in session.posargs:
                if arg:
                    args.extend(arg.split(" "))
            command.extend(args)
        session.run(*command)
    finally:
        # 0 -> OK
        # 2 -> code coverage percent unmet
        success_codes = [0, 2]

        report_cmd = ["python", "-m", "coverage", "report"]
        session.run(*report_cmd, success_codes=success_codes)

        if summary := os.getenv("GITHUB_STEP_SUMMARY"):
            report_cmd.extend(["--skip-covered", "--skip-empty", "--format=markdown"])

            with Path(summary).open("a") as output_buffer:
                output_buffer.write("")
                output_buffer.write("### Coverage\n\n")
                output_buffer.flush()
                session.run(
                    *report_cmd, stdout=output_buffer, success_codes=success_codes
                )
        else:
            session.run(
                "python",
                "-m",
                "coverage",
                "html",
                "--skip-covered",
                "--skip-empty",
                success_codes=success_codes,
            )


@nox.session
def types(session):
    session.run_install(
        "uv",
        "sync",
        "--group",
        "types",
        "--frozen",
        "--python",
        PY_LATEST,
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )

    command = ["python", "-m", "mypy", "."]
    if session.posargs:
        args = []
        for arg in session.posargs:
            if arg:
                args.extend(arg.split(" "))
        command.extend(args)
    session.run(*command)


@nox.session
def lint(session):
    session.run(
        "uv",
        "run",
        "--with",
        "pre-commit-uv",
        "--python",
        PY_LATEST,
        "pre-commit",
        "run",
        "--all-files",
    )


@nox.session
def gha_matrix(session):
    sessions = session.run("nox", "-l", "--json", silent=True)
    matrix = {
        "include": [
            {
                "python-version": session["python"],
                "django-version": session["call_spec"]["django"],
            }
            for session in json.loads(sessions)
            if session["name"] == "tests"
        ]
    }
    with Path(os.environ["GITHUB_OUTPUT"]).open("a") as fh:
        print(f"matrix={matrix}", file=fh)



from __future__ import annotations

import pytest
from django.conf import settings
from django.template.base import Template
from django.template.context import Context
from django.template.exceptions import TemplateDoesNotExist
from django.test import override_settings

from django_bird.conf import DJANGO_BIRD_BUILTINS
from django_bird.conf import DJANGO_BIRD_FINDER
from django_bird.conf import AutoConfigurator
from django_bird.conf import app_settings


@pytest.fixture(autouse=True)
def reset_settings():
    template_options = settings.TEMPLATES[0]["OPTIONS"]

    assert DJANGO_BIRD_BUILTINS in template_options["builtins"]
    assert DJANGO_BIRD_FINDER in settings.STATICFILES_FINDERS

    with override_settings(
        STATICFILES_FINDERS=[
            finder
            for finder in settings.STATICFILES_FINDERS
            if finder != DJANGO_BIRD_FINDER
        ],
        TEMPLATES=[
            settings.TEMPLATES[0]
            | {
                **settings.TEMPLATES[0],
                "OPTIONS": {
                    "builtins": [
                        builtin
                        for builtin in template_options["builtins"]
                        if builtin != DJANGO_BIRD_BUILTINS
                    ],
                },
            }
        ],
    ):
        options = settings.TEMPLATES[0]["OPTIONS"]

        assert DJANGO_BIRD_BUILTINS not in options["builtins"]
        assert DJANGO_BIRD_FINDER not in settings.STATICFILES_FINDERS

        yield


@pytest.mark.default_app_settings
def test_app_settings():
    assert app_settings.COMPONENT_DIRS == []
    assert app_settings.ENABLE_AUTO_CONFIG is True


def test_autoconfigure_disabled():
    template_options = settings.TEMPLATES[0]["OPTIONS"]

    with override_settings(
        DJANGO_BIRD={
            "ENABLE_AUTO_CONFIG": False,
        }
    ):
        app_settings.autoconfigure()

    assert DJANGO_BIRD_BUILTINS not in template_options["builtins"]
    assert DJANGO_BIRD_FINDER not in settings.STATICFILES_FINDERS


class TestAutoConfigurator:
    @pytest.fixture
    def configurator(self):
        return AutoConfigurator(app_settings)

    def test_autoconfigure(self, configurator):
        template_options = settings.TEMPLATES[0]["OPTIONS"]

        configurator.autoconfigure()

        assert DJANGO_BIRD_BUILTINS in template_options["builtins"]
        assert DJANGO_BIRD_FINDER in settings.STATICFILES_FINDERS

    def test_non_django_template_engine(self, configurator):
        template_options = settings.TEMPLATES[0]["OPTIONS"]

        with override_settings(
            TEMPLATES=[
                {
                    "BACKEND": "django.template.backends.jinja2.Jinja2",
                }
            ],
        ):
            configurator.configure_templates()

        assert DJANGO_BIRD_BUILTINS not in template_options["builtins"]

    def test_configure_builtins(self, configurator):
        template_options = settings.TEMPLATES[0]["OPTIONS"]

        configurator.configure_builtins(template_options)

        assert DJANGO_BIRD_BUILTINS in template_options["builtins"]

    def test_configure_staticfiles(self, configurator):
        configurator.configure_staticfiles()

        assert DJANGO_BIRD_FINDER in settings.STATICFILES_FINDERS

    def test_configured(self, configurator):
        assert configurator._configured is False

        configurator.autoconfigure()

        assert configurator._configured is True

    @override_settings(
        **{
            "STATICFILES_FINDERS": [
                "django.contrib.staticfiles.finders.FileSystemFinder",
                "django.contrib.staticfiles.finders.AppDirectoriesFinder",
            ],
            "TEMPLATES": [
                {
                    "BACKEND": "django.template.backends.django.DjangoTemplates",
                    "DIRS": [],
                    "APP_DIRS": True,
                    "OPTIONS": {
                        "context_processors": [
                            "django.template.context_processors.request",
                            "django.contrib.auth.context_processors.auth",
                            "django.contrib.messages.context_processors.messages",
                        ],
                    },
                },
            ],
        }
    )
    def test_startproject_settings(self, configurator, example_template):
        configurator.autoconfigure()

        template_options = settings.TEMPLATES[0]["OPTIONS"]

        assert DJANGO_BIRD_BUILTINS in template_options["builtins"]
        assert DJANGO_BIRD_FINDER in settings.STATICFILES_FINDERS

        with pytest.raises(TemplateDoesNotExist, match=example_template.base.name):
            Template(example_template.content).render(Context({}))



from __future__ import annotations

from collections.abc import Iterable
from collections.abc import Sequence
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any
from typing import Literal
from typing import final
from typing import overload

from django.contrib.staticfiles import finders
from django.contrib.staticfiles.finders import BaseFinder
from django.contrib.staticfiles.storage import StaticFilesStorage
from django.core.checks import CheckMessage
from django.core.files.storage import FileSystemStorage

from django_bird import hookimpl

from ._typing import override
from .apps import DjangoBirdAppConfig
from .conf import app_settings


class AssetType(Enum):
    CSS = "css"
    JS = "js"

    @property
    def content_type(self):
        match self:
            case AssetType.CSS:
                return "text/css"
            case AssetType.JS:
                return "application/javascript"

    @property
    def ext(self):
        return f".{self.value}"

    @classmethod
    def from_tag_name(cls, tag_name: str):
        try:
            asset_type = tag_name.split(":")[1]
            match asset_type:
                case "css":
                    return cls.CSS
                case "js":
                    return cls.JS
                case _:
                    raise ValueError(f"Unknown asset type: {asset_type}")
        except IndexError as e:
            raise ValueError(f"Invalid tag name: {tag_name}") from e


@dataclass(frozen=True, slots=True)
class Asset:
    path: Path
    type: AssetType

    @override
    def __hash__(self) -> int:
        return hash((str(self.path), self.type))

    def exists(self) -> bool:
        return self.path.exists()

    def render(self):
        if self.url is None:
            return ""

        match self.type:
            case AssetType.CSS:
                return f'<link rel="stylesheet" href="{self.url}">'
            case AssetType.JS:
                return f'<script src="{self.url}"></script>'

    @property
    def absolute_path(self):
        return self.path.resolve()

    @property
    def relative_path(self):
        return self.path.relative_to(self.template_dir)

    @property
    def storage(self):
        return BirdAssetStorage(
            location=str(self.template_dir), prefix=DjangoBirdAppConfig.label
        )

    @property
    def template_dir(self):
        template_dir = self.path.parent
        component_dirs = app_settings.get_component_directory_names()
        while (
            len(template_dir.parts) > 1 and template_dir.parts[-1] not in component_dirs
        ):
            template_dir = template_dir.parent
        return template_dir.parent

    @property
    def url(self) -> str | None:
        static_path = finders.find(str(self.relative_path))
        if static_path is None:
            return None
        static_relative_path = Path(static_path).relative_to(self.template_dir)
        return self.storage.url(str(static_relative_path))


@hookimpl
def collect_component_assets(template_path: Path) -> Iterable[Asset]:
    assets: list[Asset] = []
    for asset_type in AssetType:
        asset_path = template_path.with_suffix(asset_type.ext)
        if asset_path.exists():
            assets.append(Asset(path=asset_path, type=asset_type))
    return assets


@final
class BirdAssetStorage(StaticFilesStorage):
    def __init__(self, *args: Any, prefix: str, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.prefix = prefix


@final
class BirdAssetFinder(BaseFinder):
    def __init__(
        self, app_names: Sequence[str] | None = None, *args: Any, **kwargs: Any
    ) -> None:
        from .components import components

        self.components = components
        super().__init__(*args, **kwargs)

    @override
    def check(self, **kwargs: Any) -> list[CheckMessage]:
        return []

    # Django 5.2 changed the argument from `find` to `find_all`, but django-stubs
    # (as of the time of this commit) hasn't been updated to reflect this, hence the
    # type ignore
    @overload  # type: ignore[override]
    def find(self, path: str, *, all: Literal[False] = False) -> str | None: ...
    @overload
    def find(self, path: str, *, all: Literal[True]) -> list[str]: ...
    @overload
    def find(self, path: str, *, find_all: Literal[False] = False) -> str | None: ...
    @overload
    def find(self, path: str, *, find_all: Literal[True]) -> list[str]: ...
    @override
    def find(
        self,
        path: str,
        all: bool = False,
        find_all: bool | None = None,
    ) -> str | list[str] | None:
        """
        Given a relative file path, return the absolute path(s) where it can be found.
        """
        if find_all is None:
            find_all = all

        self.components.discover_components()

        matches: list[str] = []
        path_obj = Path(path)

        for asset in self.components.get_assets():
            if path_obj == asset.relative_path:
                matched_path = str(asset.absolute_path)
            elif asset.relative_path.is_relative_to(path_obj):
                matched_path = str(path_obj.resolve())
            else:
                continue

            if not find_all:
                return matched_path
            matches.append(matched_path)

        return matches

    @override
    def list(
        self, ignore_patterns: Iterable[str] | None
    ) -> Iterable[tuple[str, FileSystemStorage]]:
        """
        Return (relative_path, storage) pairs for all assets.
        """
        self.components.discover_components()

        for asset in self.components.get_assets():
            if ignore_patterns and any(
                asset.relative_path.match(pattern) for pattern in set(ignore_patterns)
            ):
                continue
            yield str(asset.relative_path), asset.storage



from __future__ import annotations

from django.template.loader import get_template

from .utils import TestComponent


def test_test_component(templates_dir):
    name = "foo"
    content = "<div>bar</div>"

    TestComponent(name=name, content=content).create(templates_dir)

    template = get_template(f"bird/{name}.html")

    assert template
    assert template.render({}) == content
